from typing import Any, Dict

import httpx
import pytest

from bitcoinrpc import BitcoinRPC


@pytest.mark.asyncio
async def test_connection_to_unknown_host(
    rpc_config: Dict[str, Any],
    unused_tcp_port: int,
) -> None:
    new_config = rpc_config.copy()
    new_config["url"] = f"http://localhost:{unused_tcp_port}"
    btc_rpc = BitcoinRPC.from_config(**new_config)

    with pytest.raises(httpx.ConnectError):
        await btc_rpc.getblockcount()


@pytest.mark.integration
@pytest.mark.asyncio
async def test_connection_with_incorrect_auth(rpc_config: Dict[str, Any]) -> None:
    new_config = rpc_config.copy()
    new_config["auth"] = ("a", "b")
    btc_rpc = BitcoinRPC.from_config(**new_config)

    with pytest.raises(httpx.HTTPStatusError) as e:
        await btc_rpc.getblockcount()

    # 401 - Unauthorized
    assert e.value.response.status_code == 401


@pytest.mark.integration
@pytest.mark.asyncio
async def test_connection_and_sample_rpc(rpc_config: Dict[str, Any]) -> None:
    btc_rpc = BitcoinRPC.from_config(**rpc_config)

    block_count = await btc_rpc.getblockcount()

    assert block_count >= 0


@pytest.mark.integration
@pytest.mark.asyncio
async def test_connection_and_incorrect_rpc(rpc_config: Dict[str, Any]) -> None:
    """
    Incorrect values of arguments will not raise the `bitcoinrpc.RPCError`, but server error 500.
    """
    btc_rpc = BitcoinRPC.from_config(**rpc_config)

    with pytest.raises(httpx.HTTPStatusError) as e:
        await btc_rpc.getblockheader("???")

    # 500 - Internal server error
    assert e.value.response.status_code == 500



from bitcoinrpc.__version__ import __version__
from bitcoinrpc._exceptions import RPCError
from bitcoinrpc.bitcoin_rpc import BitcoinRPC

__all__ = (
    "__version__",
    "BitcoinRPC",
    "RPCError",
)



"""
Test module exposing various errors which can occur when configuring `BitcoinRPC`.
"""

import pytest

from bitcoinrpc import BitcoinRPC


def test_unknown_httpx_option_provided_raises() -> None:
    """`httpx.AsyncClient` is strict about its init kwargs."""
    with pytest.raises(TypeError):
        BitcoinRPC.from_config(
            "http://localhost:8332",
            ("rpc_user", "rpc_passwd"),
            unknown_httpx_kwarg="foo",
        )


def test_incorrect_httpx_option_provided_raises() -> None:
    """Existing option, but incorrectly used, results in `httpx.AsyncClient` error."""
    with pytest.raises(AttributeError):
        BitcoinRPC.from_config(
            "http://localhost:8332",
            ("rpc_user", "rpc_passwd"),
            limits="Not like this.",
        )



from typing import Any, Dict, List, TypeVar, Union

from typing_extensions import Literal, NotRequired, TypedDict

JSONType = Union[None, bool, int, float, str, List["JSONType"], Dict[str, "JSONType"]]

# Type aliases for 1-to-1 match of RPC method and its return type
ConnectionCount = int
Difficulty = float
BestBlockHash = str
BlockHash = str
BlockCount = int
NetworkHashps = float
CombinePSBT = str
JoinPSBTs = str
UtxoUpdatePSBT = str


class MempoolInfo(TypedDict):
    loaded: bool
    size: int
    bytes: int
    usage: int
    maxmempoool: int
    mempoolminfee: float
    mempoolmaxfee: float


class _NetworkInfoNetworks(TypedDict):
    name: str
    limited: bool
    reachable: bool
    proxy: str
    proxy_randomize_credentials: bool


class _NetworkInfoAddresses(TypedDict):
    address: str
    port: int
    score: float


class NetworkInfo(TypedDict):
    version: int
    subversion: str
    protocolversion: str
    localservices: str
    localservicenames: List[str]
    localrelay: bool
    timeoffset: int
    networkactive: bool
    networks: List["_NetworkInfoNetworks"]
    relayfee: float
    incrementalfee: float
    localaddresses: List["_NetworkInfoAddresses"]
    warnings: str


class BlockchainInfo(TypedDict):
    # TODO: Shares common items with MiningInfo
    chain: Literal["main", "test", "regtest"]
    blocks: int
    headers: int
    bestblockhash: str
    difficulty: float
    mediantime: int
    verificationprogress: float
    initialblockdownload: bool
    chainwork: str
    size_on_disk: int
    pruned: bool
    softforks: Dict[str, Any]  # dictionary of the format {"bip-xxx": {...}}
    warnings: str


class _ChainTipsDetail(TypedDict):
    height: int
    hash: str
    branchlen: int
    status: Literal["active", "valid-fork", "valid-headers", "headers-only", "invalid"]


ChainTips = List["_ChainTipsDetail"]


class _BlockHeader(TypedDict):
    """
    Returned when verbose is set to `True`. Otherwise, `str` is returned
    """

    hash: str
    confirmations: int
    height: int
    version: int
    versionHex: str
    merkleroot: str
    time: int
    mediantime: int
    nonce: int
    bits: str
    difficulty: float
    chainwork: str
    nTx: int
    previousblockhash: str
    nextblockhash: str


BlockHeader = Union[str, "_BlockHeader"]


class BlockStats(TypedDict):
    """
    Returned dictionary will contain subset of the following, depending on filtering.
    """

    avgfee: int
    avgfeerate: int
    avgtxsize: int
    blockhash: str
    feerate_percentiles: List[int]
    heigth: int
    ins: int
    maxfee: int
    maxfeerate: int
    maxtxsize: int
    medianfee: int
    mediantime: int
    mediantxsize: int
    minfee: int
    minfeerate: int
    mintxsize: int
    outs: int
    subsidy: int  # Block reward in Satoshis
    swtotal_size: int
    swtotal_weight: int
    swtxs: int
    time: int
    total_out: int
    total_size: int
    total_weight: int
    totalfee: int
    txs: int
    utxo_increase: int
    utxo_size_inc: int


class Block(_BlockHeader):
    strippedsize: int
    tx: List["RawTransaction"]


class _RawTransaction(TypedDict):
    """Returned when verbose is set to `True`. Otherwise, `str` is returned"""

    txid: str
    hash: str
    version: int
    size: int
    vsize: int
    weight: int
    locktime: int
    vin: List[Dict[str, Any]]  # TODO: Complete
    vout: List[Dict[str, Any]]  # TODO: Complete
    hex: str
    blockhash: str
    confirmations: str
    time: int
    blocktime: int


RawTransaction = Union[str, "_RawTransaction"]


class MiningInfo(TypedDict):
    # TODO: Shares common items with BlockchainInfo
    blocks: int
    difficulty: float
    networkhashps: float
    pooledtx: int
    chain: Literal["main", "test", "regtest"]
    warnings: str


class AnalyzePSBT(TypedDict):
    inputs: List[Dict[str, Any]]  # TODO: Complete
    next: str

    fee: NotRequired[float]
    estimated_vsize: NotRequired[float]
    estimated_feerate: NotRequired[float]
    error: NotRequired[str]


class DecodePSBT(TypedDict):
    tx: Dict[str, Any]
    unknown: Dict[str, Any]
    inputs: List[Dict[str, Any]]
    outputs: List[Dict[str, Any]]

    fee: NotRequired[float]


class FinalizePSBT(TypedDict):
    psbt: str
    hex: str
    complete: bool


class WalletProcessPSBT(TypedDict):
    psbt: str
    complete: bool


BitcoinRPCResponse = TypeVar(
    "BitcoinRPCResponse",
    ConnectionCount,
    Difficulty,
    BestBlockHash,
    BlockHash,
    BlockCount,
    NetworkHashps,
    MempoolInfo,
    MiningInfo,
    BlockchainInfo,
    NetworkInfo,
    ChainTips,
    BlockHeader,
    BlockStats,
    Block,
    RawTransaction,
    CombinePSBT,
    JoinPSBTs,
    UtxoUpdatePSBT,
    AnalyzePSBT,
    DecodePSBT,
    FinalizePSBT,
    WalletProcessPSBT,
)



import itertools
from types import TracebackType
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union

import httpx
import orjson
from typing_extensions import Literal, Self

from bitcoinrpc._exceptions import RPCError
from bitcoinrpc._types import (
    AnalyzePSBT,
    BestBlockHash,
    BitcoinRPCResponse,
    Block,
    BlockchainInfo,
    BlockCount,
    BlockHash,
    BlockHeader,
    BlockStats,
    ChainTips,
    CombinePSBT,
    ConnectionCount,
    DecodePSBT,
    Difficulty,
    FinalizePSBT,
    JoinPSBTs,
    JSONType,
    MempoolInfo,
    MiningInfo,
    NetworkHashps,
    NetworkInfo,
    RawTransaction,
    UtxoUpdatePSBT,
    WalletProcessPSBT,
)

# Neat trick found in asyncio library for task enumeration
# https://github.com/python/cpython/blob/3.8/Lib/asyncio/tasks.py#L31
_next_rpc_id = itertools.count(1).__next__


class BitcoinRPC:
    __slots__ = ("_url", "_client", "_counter")
    """
    Class representing a JSON-RPC client of a Bitcoin node.

    :param url: URL of the Bitcoin node.
    :param client: Underlying `httpx.AsyncClient`, which handles the requests issued.
    :param counter: Optional callable that serves as a generator for the "id" field within JSON-RPC requests.

    For list of all available commands, visit:
    https://developer.bitcoin.org/reference/rpc/index.html
    """

    def __init__(
        self,
        url: str,
        client: httpx.AsyncClient,
        counter: Callable[[], Union[int, str]] = _next_rpc_id,
    ) -> None:
        self._url = url
        self._client = client
        self._counter = counter

    async def __aenter__(self) -> "BitcoinRPC":
        return self

    async def __aexit__(
        self,
        exc_type: Type[BaseException],
        exc_val: BaseException,
        exc_tb: TracebackType,
    ) -> None:
        await self.aclose()

    @classmethod
    def from_config(
        cls,
        url: str,
        auth: Optional[Tuple[str, str]],
        **options: Any,
    ) -> Self:
        """
        Instantiate the `BitcoinRPC` client while also configuring the underlying `httpx.AsyncClient`. Additional
        options are passed directly as kwargs to `httpx.AsyncClient`, so it's your responsibility to conform to its
        interface.
        """

        options = dict(options)
        headers = {
            "content-type": "application/json",
            **dict(options.pop("headers", {})),
        }
        return cls(url, httpx.AsyncClient(auth=auth, headers=headers, **options))

    @property
    def url(self) -> str:
        return self._url

    @property
    def client(self) -> httpx.AsyncClient:
        return self._client

    async def aclose(self) -> None:
        await self.client.aclose()

    async def acall(
        self,
        method: str,
        params: List[JSONType],
        **kwargs: Any,
    ) -> BitcoinRPCResponse:
        """
        Pass keyword arguments to directly modify the constructed request -
            see `httpx.Request`.
        """
        response = await self.client.post(
            url=self.url,
            content=orjson.dumps(
                {
                    "jsonrpc": "2.0",
                    "id": self._counter(),
                    "method": method,
                    "params": params,
                }
            ),
            **kwargs,
        )

        # Raise an exception if return code is not in 2xx range
        # https://www.python-httpx.org/quickstart/#exceptions
        response.raise_for_status()

        content = orjson.loads(response.content)
        if content["error"] is not None:
            raise RPCError(content["error"]["code"], content["error"]["message"])
        else:
            return content["result"]

    async def getmempoolinfo(self) -> MempoolInfo:
        """https://developer.bitcoin.org/reference/rpc/getmempoolinfo.html"""
        return await self.acall("getmempoolinfo", [])

    async def getmininginfo(self) -> MiningInfo:
        """https://developer.bitcoin.org/reference/rpc/getmininginfo.html"""
        return await self.acall("getmininginfo", [])

    async def getnetworkinfo(self) -> NetworkInfo:
        """https://developer.bitcoin.org/reference/rpc/getnetworkinfo.html"""
        return await self.acall("getnetworkinfo", [])

    async def getblockchaininfo(self) -> BlockchainInfo:
        """https://developer.bitcoin.org/reference/rpc/getblockchaininfo.html"""
        return await self.acall("getblockchaininfo", [])

    async def getconnectioncount(self) -> ConnectionCount:
        """https://developer.bitcoin.org/reference/rpc/getconnectioncount.html"""
        return await self.acall("getconnectioncount", [])

    async def getchaintips(self) -> ChainTips:
        """https://developer.bitcoin.org/reference/rpc/getchaintips.html"""
        return await self.acall("getchaintips", [])

    async def getdifficulty(self) -> Difficulty:
        """https://developer.bitcoin.org/reference/rpc/getdifficulty.html"""
        return await self.acall("getdifficulty", [])

    async def getbestblockhash(self) -> BestBlockHash:
        """https://developer.bitcoin.org/reference/rpc/getbestblockhash.html"""
        return await self.acall("getbestblockhash", [])

    async def getblockhash(self, height: int) -> BlockHash:
        """https://developer.bitcoin.org/reference/rpc/getblockhash.html"""
        return await self.acall("getblockhash", [height])

    async def getblockcount(self) -> BlockCount:
        """https://developer.bitcoin.org/reference/rpc/getblockcount.html"""
        return await self.acall("getblockcount", [])

    async def getblockheader(
        self, block_hash: str, verbose: bool = True
    ) -> BlockHeader:
        """https://developer.bitcoin.org/reference/rpc/getblockheader.html"""
        return await self.acall("getblockheader", [block_hash, verbose])

    async def getblockstats(
        self,
        hash_or_height: Union[int, str],
        *keys: str,
        timeout: Optional[float] = 5.0,
    ) -> BlockStats:
        """
        https://developer.bitcoin.org/reference/rpc/getblockstats.html

        Enter `keys` as positional arguments to return only the provided `keys`
            in the response.
        """
        return await self.acall(
            "getblockstats",
            [hash_or_height, list(keys) or None],
            timeout=httpx.Timeout(timeout),
        )

    async def getblock(
        self,
        block_hash: str,
        verbosity: Literal[0, 1, 2] = 1,
        timeout: Optional[float] = 5.0,
    ) -> Block:
        """
        https://developer.bitcoin.org/reference/rpc/getblock.html

        :param verbosity: 0 for hex-encoded block data, 1 for block data with
            transactions list, 2 for block data with each transaction.
        """
        return await self.acall(
            "getblock", [block_hash, verbosity], timeout=httpx.Timeout(timeout)
        )

    async def getrawtransaction(
        self,
        txid: str,
        verbose: bool = True,
        block_hash: Optional[str] = None,
        timeout: Optional[float] = 5.0,
    ) -> RawTransaction:
        """
        https://developer.bitcoin.org/reference/rpc/getrawtransactiono.html

        :param txid: If transaction is not in mempool, block_hash must also be provided.
        :param verbose: True for JSON, False for hex-encoded string
        :param block_hash: see ^txid
        :param timeout: If doing a lot of processing, no timeout may come in handy
        """
        return await self.acall(
            "getrawtransaction",
            [txid, verbose, block_hash],
            timeout=httpx.Timeout(timeout),
        )

    async def getnetworkhashps(
        self,
        nblocks: int = -1,
        height: Optional[int] = None,
        timeout: Optional[float] = 5.0,
    ) -> NetworkHashps:
        """
        https://developer.bitcoin.org/reference/rpc/getnetworkhashps.html

        :param nblocks: -1 for estimated hash power since last difficulty change,
            otherwise as an average over last provided number of blocks
        :param height: If not provided, get estimated hash power for the latest block
        :param timeout: If doing a lot of processing, no timeout may come in handy
        """
        return await self.acall(
            "getnetworkhashps", [nblocks, height], timeout=httpx.Timeout(timeout)
        )

    async def analyzepsbt(self, psbt: str) -> AnalyzePSBT:
        """
        https://developer.bitcoin.org/reference/rpc/analyzepsbt.html

        :param psbt: base64 string of a partially signed bitcoin transaction
        """
        return await self.acall("analyzepsbt", [psbt])

    async def combinepsbt(self, *psbts: str) -> CombinePSBT:
        """
        https://developer.bitcoin.org/reference/rpc/combinepsbt.html

        :param psbts: base64 strings, each representing a partially signed bitcoin transaction
        """
        return await self.acall("combinepsbt", list(psbts))

    async def decodepsbt(self, psbt: str) -> DecodePSBT:
        """
        https://developer.bitcoin.org/reference/rpc/decodepsbt.html

        :param psbt: base64 string of a partially signed bitcoin transaction
        """
        return await self.acall("decodepsbt", [psbt])

    async def finalizepsbt(self, psbt: str, extract: bool = True) -> FinalizePSBT:
        """
        https://developer.bitcoin.org/reference/rpc/finalizepsbt.html

        :param psbt: base64 string of a partially signed bitcoin transaction
        :param extract: If set to true and the transaction is complete, return a hex-encoded network transaction
        """
        return await self.acall("finalizepsbt", [psbt, extract])

    async def joinpsbts(self, *psbts: str) -> JoinPSBTs:
        """
        https://developer.bitcoin.org/reference/rpc/joinpsbts.html

        :param psbts: base64 strings, each representing a partially signed bitcoin transaction
        """
        return await self.acall("joinpsbts", list(psbts))

    async def utxoupdatepsbt(
        self,
        psbt: str,
        descriptors: Optional[List[Union[str, Dict[str, Union[int, str]]]]] = None,
    ) -> UtxoUpdatePSBT:
        """
        https://developer.bitcoin.org/reference/rpc/utxoupdatepsbt.html

        :param psbt: base64 string of a partially signed bitcoin transaction
        :param extract: If set to true and the transaction is complete, return a hex-encoded network transaction
        """
        if descriptors is not None:
            params = [psbt, descriptors]
        else:
            params = [psbt]
        return await self.acall("utxoupdatepsbt", params)  # type: ignore

    async def walletprocesspsbt(
        self,
        psbt: str,
        sign: bool = True,
        sighashtype: str = "ALL",
        bip32_derivs: bool = True,
    ) -> WalletProcessPSBT:
        """
        https://developer.bitcoin.org/reference/rpc/walletprocesspsbt.html

        :param psbt: base64 string of a partially signed bitcoin transaction
        :param sign: Sign the transaction too when updating
        :param sighashtype: signature hash type to sign, if it is not specified by PSBT.
        :param bip32_derivs: include bip32 derivation paths for pubkeys if known
        """
        return await self.acall(
            "walletprocesspsbt", [psbt, sign, sighashtype, bip32_derivs]
        )



import re
from pathlib import Path
from typing import List

from setuptools import setup


def get_version(package: str) -> str:
    """
    Extract package version, located in the `src/package/__version__.py`.
    """
    version = Path("src", package, "__version__.py").read_text()
    pattern = r"__version__ = ['\"]([^'\"]+)['\"]"
    return re.match(pattern, version).group(1)  # type: ignore


def get_requirements(req_file: str) -> List[str]:
    """
    Extract requirements from provided file.
    """
    req_path = Path(req_file)
    requirements = req_path.read_text().split("\n") if req_path.exists() else []
    return requirements


def get_long_description(readme_file: str) -> str:
    """
    Extract README from provided file.
    """
    readme_path = Path(readme_file)
    long_description = (
        readme_path.read_text(encoding="utf-8") if readme_path.exists() else ""
    )
    return long_description


setup(
    name="bitcoinrpc",
    python_requires=">=3.7",
    version=get_version("bitcoinrpc"),
    description="Lightweight Bitcoin JSON-RPC Python asynchronous client",
    long_description=get_long_description("README.md"),
    long_description_content_type="text/markdown",
    keywords="bitcoin async json-rpc",
    license="MIT",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    url="https://github.com/bibajz/bitcoin-python-async-rpc",
    author="Libor Martinek",
    author_email="libasmartinek@protonmail.com",
    package_dir={"": "src"},
    packages=["bitcoinrpc"],
    install_requires=get_requirements("requirements.txt"),
)
